PROGRAM CMmt; {Copyright 1994 Paul Milenkovic ALL RIGHTS RESERVED}

{$M 28672,0,0}

USES StrBlock, StdIO, DSPTypes, FFT, FiltBuf;


CONST
    nexp = 10;

TYPE
    power_spec = ARRAY [0 .. FFT.max_n_spec] OF SINGLE;

VAR
    cal: BOOLEAN;
    i, nspec: INTEGER;
    pstr, wstr: ASCIIZ_string;
    sample_ms, window_ms, time_step, start_ms: DOUBLE;
    scale, freq_min, freq_max: DOUBLE;

    linear_scale, bark_scale, dbark_scale, cal_spec: power_spec;
    fname, sname, cname: ASCIIZ_string;


PROCEDURE InitFreqScales(nspec: INTEGER);

    VAR
        i: INTEGER;
        freq: SINGLE;

BEGIN
    FOR i := 0 TO nspec DO BEGIN
        freq := i*0.5/(sample_ms*nspec);
        IF freq < 0.15 THEN
            freq := 0.15
        ELSE IF freq < 0.2 THEN
            freq := 0.8*freq + 0.03
        ELSE IF freq < 0.25 THEN
            freq := 1.2*freq - 0.05;
        bark_scale[i] := 13*ArcTan(0.76*freq) + 3.5*ArcTan((freq/7.5)*(freq/7.5));
        IF i > 0 THEN
            dbark_scale[i] := bark_scale[i] - bark_scale[Pred(i)]
        ELSE
            dbark_scale[i] := 0;
        linear_scale[i] := freq;
    END;
END;


PROCEDURE LinearSpectrumMoments(

    VAR spec: power_spec; imin, imax, nspec: INTEGER;
    VAR u1, u2, u3, u4: SINGLE
);

    VAR
        i: INTEGER;
        sum, moment1, moment: SINGLE;
        nsp: power_spec;

BEGIN
    sum   := 0;
    FOR i := imin TO imax DO
        sum := sum + spec[i];
    FOR i := imin TO imax DO
        nsp[i] := spec[i]/sum;

    u1    := 0;
    FOR i := imin TO imax DO
        u1 := u1 + linear_scale[i]*nsp[i];

    u2   := 0;
    u3   := 0;
    u4   := 0;
    FOR i := imin TO imax DO BEGIN
        moment1 := linear_scale[i] - u1;
        moment  := moment1*moment1;
        u2      := u2 + moment*nsp[i];
        moment  := moment*moment1;
        u3      := u3 + moment*nsp[i];
        moment  := moment*moment1;
        u4      := u4 + moment*nsp[i];
    END;
END;


PROCEDURE BarkSpectrumMoments(

    VAR spec: power_spec; imin, imax, nspec: INTEGER;
    VAR u1, u2, u3, u4: SINGLE
);

    VAR
        i: INTEGER;
        sum, moment1, moment: SINGLE;
        nsp: power_spec;

BEGIN
    sum := 0;
    FOR i := imin TO imax DO
        sum := sum + spec[i]*dbark_scale[i];
    FOR i := imin TO imax DO
        nsp[i] := spec[i]*dbark_scale[i]/sum;

    u1  := 0;
    FOR i := imin TO imax DO
        u1 := u1 + bark_scale[i]*nsp[i];

    u2  := 0;
    u3  := 0;
    u4  := 0;
    FOR i := imin TO imax DO BEGIN
        moment1 := bark_scale[i] - u1;
        moment  := moment1*moment1;
        u2      := u2 + moment*nsp[i];
        moment  := moment*moment1;
        u3      := u3 + moment*nsp[i];
        moment  := moment*moment1;
        u4      := u4 + moment*nsp[i];
    END;
END;


PROCEDURE MomentsAnalysis;

    VAR
        i, nread, nbuf, imin, imax: INTEGER;
        t, tp: DOUBLE;

        wrk_buf: FiltBuf.wrk_bufT;
        wrk_len, wrk_dif, wrk_pos: INTEGER;

        sig: DSPTypes.sig_buffer;
        frec, wind: Fourier_record;
        spec: power_spec;

        db, u1, u2, u3, u4, sigma: SINGLE;
        cal_file, moment_file: TEXT;

    PROCEDURE UpdateSamplePointer;

    BEGIN
        wrk_dif := Round((t - tp)/sample_ms);
        wrk_pos := wrk_pos + wrk_dif;
        tp      := tp + sample_ms*wrk_dif;
    END;

BEGIN
    nbuf := Round(window_ms/sample_ms);
    IF nbuf > mbuf THEN nbuf := mbuf;

    wrk_len   := 0;
    wrk_pos   := 1 - Pred(nbuf DIV 2);
    tp        := start_ms;
    t         := start_ms;
    nspec     := Twoto(Pred(nexp));

    FOR i := 0 TO nspec DO cal_spec[i] := 0;
    IF cname <> '' THEN BEGIN
        Assign(cal_file,cname);
        {$I-} Reset(cal_file) {$I+};
        IF IOResult = 0 THEN BEGIN
            FOR i := 0 TO nspec DO
                IF NOT EOF(cal_file) THEN
                    Readln(cal_file,cal_spec[i]);
            Close(cal_file);
        END ELSE BEGIN
            StdCmd(Concat('alert_msg=Cal file ',cname));
            StdCmd('alert_ack=not_found');
            Halt;
        END;
    END;

    FOR i := 0 TO Pred(nbuf) DO
        wind.s[i] := (0.54-0.46*cos((2.0*Pi*(i+0.5))/nbuf))/(0.54*nbuf);
    InitFreqScales(nspec);
    imax  := Round(2*nspec*freq_max*sample_ms);
    imin  := Round(2*nspec*freq_min*sample_ms);
    IF imax > Pred(nspec) THEN imax := Pred(nspec);
    IF imin < 1           THEN imin := 1;

    WHILE wrk_pos < 2 DO BEGIN
        t := t + time_step;
        UpdateSamplePointer;
    END;

    Assign(moment_file,fname);
    Rewrite(moment_file);

    StdCmd('');
    REPEAT
        IF wrk_len > wrk_mrk THEN BEGIN
            wrk_dif := wrk_len - wrk_mrk;
            Move(wrk_buf[Succ(wrk_dif)],wrk_buf[1],wrk_mrk*Sizeof(INTEGER));
            wrk_len := wrk_len - wrk_dif;
            wrk_pos := wrk_pos - wrk_dif;
        END;
        nread := StdBlockRead(
            wrk_buf[Succ(wrk_len)],Sizeof(wrk_buf)-Sizeof(INTEGER)*wrk_len
        );
        wrk_len := wrk_len + nread DIV Sizeof(INTEGER);

        WHILE Pred(wrk_pos + nbuf) <= wrk_len DO BEGIN
            LoadBufferValuesDif(wrk_buf,sig,wrk_pos,nbuf,scale,1.0);
            Write(moment_file,t:0:0);

            u1 := 0;
            u3 := 0;
            u4 := 0;

            dB := RMSdB(wrk_buf,wrk_pos,nbuf,scale);
            IF dB > -80.0 THEN BEGIN
                Write(moment_file,#9,dB:0:1);

                ClearFourierSignal(frec.s,nexp);
                FOR i := 0 TO Pred(nbuf) DO
                    frec.s[i] := sig[i]*wind.s[i];
                RealFourierTransform(frec,nexp);
                FOR i := 0 TO nspec DO
                    WITH frec.a[i] DO spec[i] := re*re + im*im - cal_spec[i];

                LinearSpectrumMoments(spec,imin,imax,nspec,u1,u2,u3,u4);

                IF u2 <= 0 THEN BEGIN
                    sigma := 0;
                    u3    := 0;
                    u4    := 0;
                END ELSE BEGIN
                    sigma := Sqrt(u2);
                    u3    := u3/(sigma*sigma*sigma);
                    u4    := u4/(u2*u2) - 3;
                END;
                Write(moment_file,#9,u1:0:3);
                Write(moment_file,#9,sigma:0:3);
                Write(moment_file,#9,u3:0:3);
                Write(moment_file,#9,u4:0:3);

                BarkSpectrumMoments(spec,imin,imax,nspec,u1,u2,u3,u4);

                IF u2 <= 0 THEN BEGIN
                    sigma := 0;
                    u3    := 0;
                    u4    := 0;
                END ELSE BEGIN
                    sigma := Sqrt(u2);
                    u3    := u3/(sigma*sigma*sigma);
                    u4    := u4/(u2*u2) - 3;
                END;
                Write(moment_file,#9,u1:0:3);
                Write(moment_file,#9,sigma:0:3);
                Write(moment_file,#9,u3:0:3);
                Write(moment_file,#9,u4:0:3);

                Write(moment_file,' t=',t:3:0);
                Write(moment_file,' ');
                Writeln(moment_file,sname);
            END ELSE BEGIN
                Write(moment_file,#9,u1:0:3);
                Write(moment_file,#9,u1:0:3);
                Write(moment_file,#9,u3:0:3);
                Write(moment_file,#9,u4:0:3);
                Write(moment_file,#9,u1:0:3);
                Write(moment_file,#9,u1:0:3);
                Write(moment_file,#9,u3:0:3);
                Write(moment_file,#9,u4:0:3);
                Write(moment_file,' t=',t:3:0);
                Write(moment_file,' ');
                Writeln(moment_file,sname);
            END;

            t  := t + time_step;
            UpdateSamplePointer;
        END;
    UNTIL nread = 0;

    Close(moment_file);
END;


PROCEDURE MomentsCal;

    VAR
        i, nread, nbuf, count: INTEGER;
        t, tp: DOUBLE;

        wrk_buf: FiltBuf.wrk_bufT;
        wrk_len, wrk_dif, wrk_pos: INTEGER;

        sig: DSPTypes.sig_buffer;
        frec, wind: Fourier_record;

        cal_file: TEXT;

    PROCEDURE UpdateSamplePointer;

    BEGIN
        wrk_dif := Round((t - tp)/sample_ms);
        wrk_pos := wrk_pos + wrk_dif;
        tp      := tp + sample_ms*wrk_dif;
    END;

BEGIN
    nbuf := Round(window_ms/sample_ms);
    IF nbuf > mbuf THEN nbuf := mbuf;

    wrk_len   := 0;
    wrk_pos   := 1 - Pred(nbuf DIV 2);
    tp        := start_ms;
    t         := start_ms;
    nspec     := Twoto(Pred(nexp));

    FOR i := 0 TO nspec DO cal_spec[i] := 0;

    FOR i := 0 TO Pred(nbuf) DO
        wind.s[i] := (0.54-0.46*cos((2.0*Pi*(i+0.5))/nbuf))/(0.54*nbuf);

    WHILE wrk_pos < 2 DO BEGIN
        t := t + time_step;
        UpdateSamplePointer;
    END;

    count := 0;

    StdCmd('');
    REPEAT
        IF wrk_len > wrk_mrk THEN BEGIN
            wrk_dif := wrk_len - wrk_mrk;
            Move(wrk_buf[Succ(wrk_dif)],wrk_buf[1],wrk_mrk*Sizeof(INTEGER));
            wrk_len := wrk_len - wrk_dif;
            wrk_pos := wrk_pos - wrk_dif;
        END;
        nread := StdBlockRead(
            wrk_buf[Succ(wrk_len)],Sizeof(wrk_buf)-Sizeof(INTEGER)*wrk_len
        );
        wrk_len := wrk_len + nread DIV Sizeof(INTEGER);

        WHILE Pred(wrk_pos + nbuf) <= wrk_len DO BEGIN
            LoadBufferValuesDif(wrk_buf,sig,wrk_pos,nbuf,scale,1.0);

            IF RMSdB(wrk_buf,wrk_pos,nbuf,scale) > -80.0 THEN BEGIN
                Inc(count);

                ClearFourierSignal(frec.s,nexp);
                FOR i := 0 TO Pred(nbuf) DO
                    frec.s[i] := sig[i]*wind.s[i];
                RealFourierTransform(frec,nexp);
                FOR i := 0 TO nspec DO WITH frec.a[i] DO
                    cal_spec[i] := cal_spec[i] + re*re + im*im;
            END;

            t  := t + time_step;
            UpdateSamplePointer;
        END;
    UNTIL nread = 0;

    IF count > 0 THEN BEGIN
        Assign(cal_file,cname);
        Rewrite(cal_file);
        FOR i := 0 TO nspec DO Writeln(cal_file,cal_spec[i]/count:0:20);
        Close(cal_file);
    END;
END;


BEGIN
    StdSetErrorExit;
    StdCmd('extension=MMT');
    StdCmd('leader');
    StdCmd('input');
    sample_ms := StdDbl('sample_ms?');
    start_ms  := StdDbl('start_ms?');
    scale     := Sqrt(2) * StdDbl('scale?');
    fname     := StdTxt('fname?');
    sname     := StdTxt('sname?');

    cal       := FALSE;
    window_ms := 0;
    time_step := 0;
    freq_min  := 0;
    freq_max  := 0;
    cname     := '';

    FOR i := 1 TO ParamCount DO BEGIN
        pstr := ParamStr(i);
        IF pstr = '/C' THEN cal := TRUE;
        UpdateDoubleKeyword(window_ms,pstr,'/W:');
        UpdateDoubleKeyword(time_step,pstr,'/T:');
        UpdateDoubleKeyword(freq_min, pstr,'/I:');
        UpdateDoubleKeyword(freq_max, pstr,'/F:');
        IF Copy(pstr,1,1) <> '/' THEN cname := pstr;
    END;

    IF window_ms = 0 THEN window_ms := 20;
    IF freq_max  = 0 THEN freq_max  := 0.5/sample_ms;

    IF window_ms > mbuf*sample_ms THEN BEGIN
        Str(mbuf*sample_ms:0:3,wstr);
        window_ms := StdDbl(Concat(
            'DOUBLE:Window (max = ',wstr,' ms)'));
    END;
    IF time_step = 0 THEN
        time_step := window_ms/2;
    IF window_ms > mbuf*sample_ms THEN
        window_ms := mbuf*sample_ms;
    IF window_ms > 0 THEN BEGIN
        IF cal THEN
            MomentsCal
        ELSE
            MomentsAnalysis;
    END;
END.
